import asyncio
import json
import random
import ssl
from enum import Enum
from typing import AsyncGenerator, Callable

import requests
import websockets

from pyqqq.brokerage.ebest.oauth import EBestAuth
from pyqqq.utils.logger import get_logger
from pyqqq.utils.market_schedule import *
from pyqqq.utils.retry import retry


class EBestTRClient:
    logger = get_logger(__name__ + ".EBestTRClient")

    """
    한국투자증권 TR 요청을 위한 클라이언트

    Args:
        auth (KISAuth): 인증 정보를 담고 있는 객체
        corp_data (dict): 기업 고객의 경우 추가로 필요한 정보를 담고 있는 객체
    """

    def __init__(self, auth: EBestAuth, corp_data: dict = None):
        self.auth = auth
        self.corp_data = corp_data
        self.session = requests.Session()

    @retry(requests.HTTPError, delay=1)
    def request(self, path: str, tr_cd: str, tr_cont: str = "N", tr_cont_key: str = "", params: dict = None, body: dict = None, method: str = "POST"):
        """
        TR 요청을 보내고 응답을 받는 메서드

        Args:
            path (str): 요청을 보낼 URL 경로
            tr_cd (str): TR ID
            tr_cont (str): TR CONT
            params (dict): URL 쿼리 파라미터
            body (dict): 요청 바디
            method (str): HTTP 메서드

        Returns:
            tuple: 응답 바디와 응답 헤더를 담은 튜플

        Raises:
            requests.HTTPError: 요청이 실패한 경우
        """
        headers = {
            "content-type": "application/json; charset=utf-8",
            "authorization": f"Bearer {self.auth.get_token()}",
            "tr_cd": tr_cd,
            "tr_cont": tr_cont,
            "tr_cont_key": tr_cont_key,
        }
        if self.corp_data is not None:
            headers.update(self.corp_data)

        url = f"{self.auth.host_url}{path}"

        r = self.session.request(method=method, url=url, headers=headers, params=params, json=body)

        if r.status_code != 200:
            try:
                res_body = r.json()
                if "rsp_cd" in res_body and res_body["rsp_cd"] == "IGW00121":
                    self.auth.get_token(refresh=True)

            except Exception as e:
                self.logger.exception(e)

            print(r.text)
            r.raise_for_status()

        response_headers = r.headers
        response_body = r.json()

        return response_body, response_headers


class EBestTRWebsocketKeepConnectionStatus(Enum):
    """
    LS(구 이베스트투자)증권 TR 요청 웹소켓 클라이언트의 연결 유지 상태
    """

    WAIT = "wait"
    """ 대기 """
    KEEP = "keep"
    """ 유지 """
    CLOSE = "close"
    """ 종료 """


class EBestTRWebsocketClient:
    """
    LS(구 이베스트투자)증권 실시간 데이터 요청을 위한 웹소켓 클라이언트

    Args:
        auth (KISAuth): 인증 정보를 담고 있는 객체
        on_connect (Callable): 웹소켓 연결 시 호출할 콜백 함수
        on_ask_keep_connection (Callable): 연결 유지 여부를 묻는 메시지를 받았을 때 호출할 콜백 함수. 'wait', 'keep', 'close' 중 하나를 반환해야 함
        stop_event (asyncio.Event): 종료 이벤트
    """

    logger = get_logger(__name__ + ".EBestTRWebsocketClient")

    def __init__(
        self,
        auth: EBestAuth,
        on_connect: Callable[[websockets.WebSocketClientProtocol], asyncio.Task] = None,
        on_ask_keep_connection: Callable[[], EBestTRWebsocketKeepConnectionStatus] = lambda: EBestTRWebsocketKeepConnectionStatus.KEEP,
        stop_event: asyncio.Event = None,
    ):

        self.auth = auth
        self.task = None
        self.on_connect = on_connect
        self.on_ask_keep_connection = on_ask_keep_connection
        self.stop_event = stop_event or asyncio.Event()
        self.session_id = self.make_session_id()

    def make_session_id(self) -> str:
        """세션 ID 생성

        Returns:
            영문자 + 숫자 형식 12자리 문자열
        """
        return "".join(random.choices("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789", k=12))

    async def listen(self) -> AsyncGenerator:
        """
        웹소켓을 통해 실시간 데이터를 수신하는 메서드

        Yields:
            dict: 실시간 데이터

        """
        websocket_url = f"{self.auth.websocket_url}/websocket"

        while not self.stop_event.is_set():
            try:
                keep_connection = self.on_ask_keep_connection()

                if keep_connection == EBestTRWebsocketKeepConnectionStatus.CLOSE:
                    break

                elif keep_connection == EBestTRWebsocketKeepConnectionStatus.WAIT:
                    await asyncio.sleep(0.5)
                    continue

                async with websockets.connect(websocket_url) as ws:
                    self.logger.info(f"{self.session_id}: connected to websocket")
                    if self.on_connect:
                        await self.on_connect(ws)

                    while not self.stop_event.is_set():
                        data = await self.recv_with_timeout(ws, timeout=1)
                        if data:
                            json_data = json.loads(data)
                            if json_data.get("body") is not None:
                                yield json_data

                            else:
                                if json_data["header"]["rsp_cd"] != "00000":
                                    self.logger.warn(json_data)

                        elif self.on_ask_keep_connection() == EBestTRWebsocketKeepConnectionStatus.CLOSE:
                            break

                        await asyncio.sleep(0.01)

            except websockets.exceptions.ConnectionClosedError:
                self.logger.error(f"{self.session_id}: ConnectionClosedError")
                await asyncio.sleep(0.5)
                continue

            except websockets.ConnectionClosed as e:
                await asyncio.sleep(0.5)
                self.logger.exception(e)
                continue

            except ssl.SSLZeroReturnError:
                self.logger.error(f"{self.session_id}: SSLZeroReturnError")
                await asyncio.sleep(0.5)
                continue

            except TimeoutError as e:
                self.logger.exception(e)
                await asyncio.sleep(1)
                continue

            except Exception as e:
                # any other exception
                self.logger.exception(e)
                await asyncio.sleep(1)
                continue

    def is_safe_to_close(self) -> bool:
        now = dtm.datetime.now()
        today = now.date()
        schedule = get_market_schedule(today)
        if schedule.full_day_closed:
            return True

        close_time = dtm.datetime.combine(today, schedule.close_time)
        safe_close_time = close_time + dtm.timedelta(hours=2, minutes=35)
        return now > safe_close_time

    def is_before_opening(self) -> bool:
        now = dtm.datetime.now()
        today = now.date()
        schedule = get_market_schedule(today)
        if schedule.full_day_closed:
            return True

        open_time = dtm.datetime.combine(today, schedule.open_time)
        safe_listen_time = open_time - dtm.timedelta(minutes=31)
        return now < safe_listen_time

    async def recv_with_timeout(self, ws: websockets.WebSocketClientProtocol, timeout: int = 1) -> str:
        """timeout 초 동안 수신을 대기하고 수신이 없으면 None을 반환한다"""
        try:
            message = await asyncio.wait_for(ws.recv(), timeout=timeout)
            return message
        except asyncio.TimeoutError:
            return None

    def stop(self):
        if not self.stop_event.is_set():
            self.stop_event.set()



import datetime as dtm
from typing import Optional, Tuple
from functools import lru_cache

import pandas as pd
from pyqqq.data import domestic
from pyqqq.utils.market_schedule import is_full_day_closed, get_last_trading_day
from pyqqq.utils.logger import get_bare_logger as get_logger


class DailyTickers:
    _instance = None
    _initialized = False
    logger = get_logger(__name__ + '.DailyTickers')

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        if self._initialized:
            return
        self._initialized = True

        self._today = dtm.date.today()
        if is_full_day_closed(self._today):
            self._date = get_last_trading_day(self._today)
        else:
            self._date = self._today
        self._is_before_today_krx_open = True
        self._tickers = None
        self._change_date()

    def _chk_days_passed(self):
        return self._today != dtm.date.today()

    @lru_cache(maxsize=30)  # memory_profiler 로 확인 결과 하루치 fetch 결과가 약 2.5MiB
    @staticmethod
    def fetch_tickers(date: dtm.date, exchange: str) -> pd.DataFrame:
        DailyTickers.logger.debug(f'\tfetch_tickers date={date} cache_info={DailyTickers.fetch_tickers.cache_info()}')
        return domestic.get_tickers(date, exchange=exchange)

    def _change_date(self, date: Optional[dtm.date] = None, force: bool = False):
        """_tickers 가 비었거나 날짜가 바뀌었으면 새로 채워넣는다. 8시 50분 이전에는 NXT 거래소 데이터를 사용한다."""
        if not date:
            if self._chk_days_passed():
                self._today = dtm.date.today()
                if is_full_day_closed(self._today):
                    date = get_last_trading_day(self._today)
                else:
                    date = self._today
            else:
                date = self._date

        if self._today == date:
            if dtm.datetime.now().time() < dtm.time(8, 50) and not self._is_before_today_krx_open:
                self._is_before_today_krx_open = True
                self._tickers = None
            elif dtm.datetime.now().time() >= dtm.time(8, 50) and self._is_before_today_krx_open:
                self._is_before_today_krx_open = False
                self._tickers = None
            else:
                pass
        else:
            self._is_before_today_krx_open = False

        if force:
            DailyTickers.fetch_tickers.cache_clear()
            self._tickers = None

        if self._tickers is None or self._date != date:
            self._date = date
            exchange = "NXT" if self._is_before_today_krx_open else "KRX"
            self.logger.debug(f'\tdate changed. date={self._date} wait for get_tickers()')
            self._tickers = DailyTickers.fetch_tickers(self._date, exchange=exchange)

    def get_tickers(self, date: Optional[dtm.date] = None, force: bool = False) -> pd.DataFrame:
        """
        종목정보 가져오기
        """
        self._change_date(date, force)
        return self._tickers

    def get_ticker_info(self, code: str, date: Optional[dtm.date] = None, force: bool = False) -> Tuple[Optional[str], Optional[str]]:
        """
        종목정보 가져오기
        """
        self._change_date(date, force)

        # self.logger.debug(f'\tget_ticker_info code={code} date={self._date}')
        try:
            name = self._tickers.loc[code, 'name']
            type = self._tickers.loc[code, 'type']
        except KeyError:
            try:
                self._change_date(date, True)
                name = self._tickers.loc[code, 'name']
                type = self._tickers.loc[code, 'type']
            except KeyError:
                self.logger.warning(f'KeyError on get_ticker_info. code={code}')
            return (None, None)

        return (name, type)

    def get_listing_date(self, code: str, date: Optional[dtm.date] = None, force: bool = False) -> Optional[dtm.date]:
        """
        상장일 가져오기
        """
        self._change_date(date, force)

        try:
            list_date = self._tickers.loc[code, 'listing_date']
        except KeyError:
            try:
                self._change_date(date, True)
                list_date = self._tickers.loc[code, 'listing_date']
            except KeyError:
                self.logger.warning(f'KeyError on get_listing_date. code={code}')
            return None

        return dtm.datetime.strptime(list_date, "%Y%m%d").date() if list_date else None

    def get_ticker_upper_limit(self, code: str, date: Optional[dtm.date] = None, force: bool = False) -> Optional[int]:
        """
        상한가 가져오기
        """
        self._change_date(date, force)

        # self.logger.debug(f'\tget_ticker_upper_limit code={code} date={self._date}')
        try:
            upper_limit = self._tickers.loc[code, 'upper_limit']
        except KeyError:
            try:
                self._change_date(date, True)
                upper_limit = self._tickers.loc[code, 'upper_limit']
            except KeyError:
                self.logger.warning(f'KeyError on get_ticker_upper_limit. code={code}')
                return None

        return int(upper_limit) if upper_limit else None



from dotenv import load_dotenv
from pyqqq.brokerage.ebest.simple import EBestSimpleDomesticStock
from pyqqq.brokerage.kis.simple import KISSimpleDomesticStock
from pyqqq.datatypes import *
from pyqqq.utils.kvstore import KVStore
from pyqqq.utils.logger import get_logger
from pyqqq.brokerage.tracker import TradingTracker


load_dotenv()


class PositionClassifier:
    """주문, 포지션의 주체를 분류하기 위한 클래스입니다.

    자동 거래 프로그램에서 사용중인 계좌로 직접 HTS, MTS 등으로 거래를 하는 유저를 위해 만들어진 분류기 입니다.
    """

    logger = get_logger(__name__ + ".PositionClassifier")

    DEF_DIRECT_POSITION_KEY = "direct_position"
    DEF_DIRECT_ORDER_KEY = "direct_order"
    DEF_AUTO_POSITION_KEY = "auto_position"
    DEF_AUTO_ORDER_KEY = "auto_order"
    DEF_TAG_AUTO_ORDER_KEY = "tag_auto_order"

    def __init__(
        self,
        simple_data_api: EBestSimpleDomesticStock | KISSimpleDomesticStock,
        kv_store_collection,
        default_type="auto",
    ):
        """OrderClassifier 초기화 메서드입니다

        Args:
            data_api (KISSimpleDomesticStock):
                시세 조회 및 시장 데이터 조회를 위한 API 인터페이스
            kv_store_collection (String):
                kv_store에서 사용할 콜렉션
            default_type (String):
                태깅되지 않은 종목을 무엇으로 간주할지.
                'auto', 'direct' 둘 중 한 값을 가짐

        Note:
            생성된 인스턴스는 즉시 거래가 가능한 상태가 되며,
            모든 거래 관련 작업은 자동으로 로깅됩니다.

        """

        # key: asset_code(str), value: quantity(int)
        self.auto_positions = {}
        self.direct_positions = {}

        # key: order_no(str), value: remain_quantity(int)
        self.auto_pending_orders = {}
        self.direct_pending_orders = {}

        # list of order_no(str)
        self.tagged_auto_pending_orders = []

        self.kv_store = KVStore(kv_store_collection)
        self.api = simple_data_api

        if default_type in ['auto', 'direct']:
            self.default_type = default_type
        else:
            self.default_type = 'auto'

        self.tracker = TradingTracker(simple_data_api)
        self.tracker.add_pending_order_update_callback(self._on_pending_order_update)

        self.set_initial_position()

    def set_initial_position(self) -> None:
        """코드 시작시 현재 account에 있는 position과 kv_store를 참고해서
        기존 포지션을 분류한다.
        Args:
        """
        cur_pos = self.api.get_positions()
        kv_auto_pos = self.kv_store.get(self.DEF_AUTO_POSITION_KEY) or {}
        kv_direct_pos = self.kv_store.get(self.DEF_DIRECT_POSITION_KEY) or {}

        self.logger.info(f"set_initial_position. kv_auto_pos: {kv_auto_pos}, kv_direct_pos: {kv_direct_pos}")
        for pos in cur_pos:
            asset_code = pos.asset_code
            quantity = pos.quantity

            if kv_auto_pos is not None and asset_code in kv_auto_pos:
                self.auto_positions[asset_code] = min([quantity, kv_auto_pos[asset_code]])
                quantity -= self.auto_positions[asset_code]
            if asset_code in kv_direct_pos:
                self.direct_positions[asset_code] = min(quantity, kv_direct_pos[asset_code])
                quantity -= self.direct_positions[asset_code]
            if quantity > 0:
                if self.default_type == "auto":
                    if asset_code in self.auto_positions:
                        self.auto_positions[asset_code] += quantity
                    else:
                        self.auto_positions[asset_code] = quantity
                elif self.default_type == "direct":
                    if asset_code in self.direct_positions:
                        self.direct_positions[asset_code] += quantity
                    else:
                        self.direct_positions[asset_code] = quantity

        self.update_positions()
        self.logger.info(f"set_initial_position.\npositions: {cur_pos}\nauto_positions: {self.auto_positions}\ndirect_positions: {self.direct_positions}")

    def set_initial_order(self):
        cur_order = self.api.get_pending_orders(exchanges=[OrderExchange.KRX, OrderExchange.NXT, OrderExchange.SOR])
        kv_auto_orders = self.kv_store.get(self.DEF_AUTO_ORDER_KEY) or {}
        kv_direct_orders = self.kv_store.get(self.DEF_DIRECT_ORDER_KEY) or {}
        kv_auto_tag_orders = self.kv_store.get(self.DEF_TAG_AUTO_ORDER_KEY) or []

        self.logger.info(f"set_initial_order. kv_auto_orders: {kv_auto_orders}, kv_direct_orders: {kv_direct_orders}")

        for order in cur_order:
            order_no = cur_order.order_no
            quantity = cur_order.pending_quantity

            if order_no in kv_auto_tag_orders:
                self.tagged_auto_pending_orders.append(order_no)

            if order_no in kv_auto_orders:
                self.auto_pending_orders[order_no] = min(quantity, kv_auto_orders[order_no])
            elif order_no in kv_direct_orders:
                self.direct_pending_orders[order_no] = min(quantity, kv_direct_orders[order_no])
            else:
                if self.default_type == "auto":
                    self.auto_pending_orders[order_no] = quantity
                elif self.default_type == "direct":
                    self.direct_pending_orders[order_no] = quantity

        self.update_orders()
        self.logger.info(f"set_initial_order.\orders: {cur_order} \nauto_pending_orders: {self.auto_pending_orders} \ndirect_pending_orders: {self.direct_pending_orders}")

    def update_positions(self) -> None:
        self.kv_store.set(self.DEF_AUTO_POSITION_KEY, self.auto_positions)
        self.kv_store.set(self.DEF_DIRECT_POSITION_KEY, self.direct_positions)

    def update_orders(self):
        self.kv_store.set(self.DEF_AUTO_ORDER_KEY, self.auto_pending_orders)
        self.kv_store.set(self.DEF_TAG_AUTO_ORDER_KEY, self.tagged_auto_pending_orders)
        self.kv_store.set(self.DEF_DIRECT_ORDER_KEY, self.direct_pending_orders)

    def clear_orders(self) -> None:
        """
        날 바뀌고 kv store에 남아있는 오더는 모두 유효지 않음
        장 시작 전에 호출해주면 좋은 함수
        """
        self.logger.info("clear_orders")
        self.auto_pending_orders = {}
        self.direct_pending_orders = {}
        self.tagged_auto_pending_orders = []
        self.update_orders()

    def get_sellable_quantity_by_auto(self, asset_code, quantity) -> int:
        """pyqqq로 매수했던 수량을 체크해서, 매도간에 pyqqq로 샀던게 몇 주인지 확인하는 함수

        Args:
            asset_code (String): 종목코드 6자리
            quantity (int): 체크 하려는 수량.
            만약 프로그램으로 5주, 손으로 5주 매수해서 10주 들고있는 상황에서 quantity가 8이면
            5를 return 한다.
        """
        ret = 0

        positions = self.api.get_positions()
        for position in positions:
            if asset_code == position.asset_code:
                remained = min(quantity, position.sell_possible_quantity)
                if asset_code in self.auto_positions:
                    diff = min(self.auto_positions[asset_code], remained)
                    ret += diff
                    remained -= diff

                if self.default_type == "auto":
                    if asset_code in self.direct_positions:
                        diff = min(self.direct_positions[asset_code], remained)
                        remained -= diff

                    ret += remained
                break
        return ret

    # 매수, 매도 할때 명시적으로 호출해줘야함
    def tagging_order_auto(self, order_no: str):
        self.logger.info(f"tagging_order_auto. order_no: {order_no}")
        if order_no in self.auto_pending_orders:
            self.logger.info(f"tagging_order_auto. {order_no} already exist")
            return

        elif order_no in self.direct_pending_orders:
            self.logger.info(f"tagging_order_auto. {order_no} has been in direct_pending_orders")
            self.auto_pending_orders[order_no] = self.direct_pending_orders[order_no]

            del self.direct_pending_orders[order_no]

        elif order_no not in self.tagged_auto_pending_orders:
            self.logger.info(f"tagging_order_auto. {order_no} is tagged.")
            self.tagged_auto_pending_orders.append(order_no)

        self.update_orders()

    def _on_pending_order_update(self, status, order: StockOrder):
        """Tracker를 통해
        TradingTracker의 add_pending_order_update_callback 함수에 인자로 넣으면 정상 동작함.
        """
        asset_code = order.asset_code
        order_no = order.order_no
        pending_quantity = order.pending_quantity

        self.logger.info(f"_on_pending_order_update. status: {status} order: {order}")

        # 주문 처리시
        if order.side == OrderSide.BUY:
            if status in ["partial", "completed"]:
                if order_no in self.auto_pending_orders:
                    prev_filled_quantity = order.quantity - self.auto_pending_orders[order_no]
                    filled_quantity = order.filled_quantity - prev_filled_quantity
                    self._add_dict_with_key(self.auto_pending_orders, order_no, -filled_quantity)
                    self._add_dict_with_key(self.auto_positions, asset_code, filled_quantity)

                elif order_no in self.direct_pending_orders:
                    prev_filled_quantity = order.quantity - self.direct_pending_orders[order_no]
                    filled_quantity = order.filled_quantity - prev_filled_quantity
                    self._add_dict_with_key(self.direct_pending_orders, order_no, -filled_quantity)
                    self._add_dict_with_key(self.direct_positions, asset_code, filled_quantity)

                elif order_no in self.tagged_auto_pending_orders or self.default_type == "auto":
                    self.auto_pending_orders[order_no] = pending_quantity
                    self._add_dict_with_key(self.auto_positions, asset_code, order.filled_quantity)

                elif self.default_type == "direct":
                    self.direct_pending_orders[order_no] = pending_quantity
                    self._add_dict_with_key(self.direct_positions, asset_code, order.filled_quantity)

            # 주문 접수시
            elif status in ["accepted"]:
                if order_no in self.tagged_auto_pending_orders:
                    self.auto_pending_orders[order_no] = order.pending_quantity
                else:
                    self.direct_pending_orders[order_no] = order.pending_quantity

            elif status in ["cancelled"]:
                if order_no in self.auto_pending_orders:
                    self.auto_pending_orders[order_no] -= order.quantity
                elif order_no in self.direct_pending_orders:
                    self.direct_pending_orders[order_no] -= order.quantity

        elif order.side == OrderSide.SELL:
            if status in ["accepted"]:
                # 매도시엔 보유 잔량을 주문 접수시 미리 차감한다.
                if order_no in self.tagged_auto_pending_orders:
                    self.auto_pending_orders[order_no] = pending_quantity

                    remain = self._add_dict_with_key(self.auto_positions, asset_code, -pending_quantity)

                    if remain < 0:
                        self._add_dict_with_key(self.direct_positions, asset_code, remain)

                else:
                    self.direct_pending_orders[order_no] = pending_quantity
                    remain = self._add_dict_with_key(self.direct_positions, asset_code, -pending_quantity)

                    if remain < 0:
                        self._add_dict_with_key(self.auto_positions, asset_code, remain)

            elif status in ["cancelled"]:
                # 주문 취소시엔 보유잔량을 복구한다.
                if order_no in self.auto_pending_orders:
                    # order.quantity가 주문 취소한 수량임
                    self.auto_pending_orders[order_no] -= order.quantity
                    self._add_dict_with_key(self.auto_positions, asset_code, order.quantity)
                elif order_no in self.direct_pending_orders:
                    self.direct_pending_orders[order_no] -= order.quantity
                    self._add_dict_with_key(self.direct_positions, asset_code, order.quantity)
            elif status in ["partial"]:
                # 부분 체결시엔 미체결분량으로 pending order 잔량을 업데이트한다.
                if order_no in self.auto_pending_orders:
                    self.auto_pending_orders[order_no] = pending_quantity
                elif order_no in self.direct_pending_orders:
                    self.direct_pending_orders[order_no] = pending_quantity
            elif status in ["completed"]:
                # 전체 체결시엔 메모리 정리를 위해 해당 주문번호를 비워준다. 진행하지 않아도 무방함
                if order_no in self.auto_pending_orders:
                    del self.auto_pending_orders[order_no]
                elif order_no in self.direct_pending_orders:
                    del self.direct_pending_orders[order_no]

        # clear
        self._clear_below_zero_key_value(self.auto_pending_orders, order_no)
        self._clear_below_zero_key_value(self.direct_pending_orders, order_no)
        self._clear_below_zero_key_value(self.auto_positions, asset_code)
        self._clear_below_zero_key_value(self.direct_positions, asset_code)

        self.update_orders()
        self.update_positions()

    def print_current_status(self):
        self.logger.info(
            f"print_current_status. auto positions: {self.auto_positions} direct positions: {self.direct_positions}.\nauto orders: {self.auto_pending_orders} direct_orders: {self.direct_pending_orders}.\n tagged order_nos: {self.tagged_auto_pending_orders}"
        )

    async def start(self):
        return await self.tracker.start()

    def _add_dict_with_key(self, target_dict, key, diff):
        """
        target dict가 비어있든 아니든 적절히 diff를 더해주는 연산.
        만약 연산 결과가 음수라면 그 음수값을 return하여 적절히 사용할 수 있도록 함
        """
        if key in target_dict:
            target_dict[key] += diff
        else:
            target_dict[key] = diff

        if target_dict[key] < 0:
            return target_dict[key]
        else:
            return 0

    def _clear_below_zero_key_value(self, target_dict, key):
        if key in target_dict and target_dict[key] <= 0:
            del target_dict[key]



import datetime
from typing import Dict, List, Optional, Union

import pandas as pd
import pytz

import pyqqq.config as c
from pyqqq.datatypes import DataExchange
from pyqqq.utils.api_client import raise_for_status, send_request
from pyqqq.utils.array import chunk
from pyqqq.utils.local_cache import DiskCacheManager
from pyqqq.utils.logger import get_logger

logger = get_logger(__name__)
dailyCache = DiskCacheManager("daily_cache")


@dailyCache.memoize()
def get_all_ohlcv_for_date(
    date: datetime.date,
    adjusted: bool = True,
    exchange: Union[str, DataExchange] = "KRX",
) -> pd.DataFrame:
    """
    주어진 날짜에 대한 모든 주식의 OHLCV(Open, High, Low, Close, Volume) 데이터를 조회합니다.

    이 함수는 특정 날짜에 대한 모든 주식의 시가, 고가, 저가, 종가 및 거래량 데이터를 API를 통해 요청하고,
    이를 pandas DataFrame 형태로 반환합니다. 반환된 DataFrame은 'code'를 인덱스로 사용합니다.

    KRX: 2018년 1월 1일 데이터 부터 조회 가능합니다.
    NXT: 2025년 3월 4일 데이터 부터 조회 가능합니다.

    Args:
        date (datetime.date): 조회할 날짜.
        adjusted (bool): 수정주가 여부. 기본값은 True.
        exchange (Union[str, DataExchange]): 거래소. 기본값은 KRX.

    Returns:
        pd.DataFrame: OHLCV 데이터를 포함하는 DataFrame. 'code' 컬럼은 DataFrame의 인덱스로 설정됩니다.

        DataFrame의 컬럼은 다음과 같습니다.

        - open (int): 시가.
        - high (int): 고가.
        - low (int): 저가.
        - close (int): 종가.
        - volume (int): 거래량.
        - value (int): 거래대금.

    Raises:
        HTTPError: API 요청이 실패했을 때 발생.

    Examples:
        >>> ohlcv_data = get_all_ohlcv_for_date(datetime.date(2025, 3, 4), exchange="KRX")
        >>> print(ohlcv_data)
                    open   high    low  close  volume        value
            code
            000020   6200   6220   6090   6130   41954    257785065
            000040    422    422    393    400  135979     54618038
            000050   6350   6350   6240   6290    9957     62588010
            000070  59800  61100  59500  60400   32555   1957081250
            000075  59000  59400  58300  59400    2600    151987900
            ...       ...    ...    ...    ...     ...          ...
            950160  35500  39150  35500  38950  332462  12638215825
            950170   4900   4955   4880   4900   24198    118688025
            950190   7860   7920   7770   7860   21309    166408215
            950200   3855   3855   3785   3800   23268     88458215
            950220    996    996    945    954  440017    421451605
    """
    if isinstance(date, datetime.datetime):
        date = date.date()

    exchange = DataExchange.validate(exchange)
    url = f"{c.PYQQQ_API_URL}/domestic-stock/ohlcv/daily/all/{date}"
    r = send_request(
        "GET",
        url,
        params={
            "adjusted": "true" if adjusted else "false",
            "current_date": datetime.date.today(),
            "exchange": exchange.value,
        },
    )
    raise_for_status(r)

    data = r.json()

    cols = data["cols"]
    rows = data["rows"]

    if not rows:
        return pd.DataFrame()
    else:
        df = pd.DataFrame(rows, columns=cols)

        # NOTE 서버 응답에서 diff, diff_rate 가 제거될 예정
        columns_to_drop = ["diff", "diff_rate", "date"]
        df.drop(columns=[col for col in columns_to_drop if col in df.columns], inplace=True)
        df.set_index("code", inplace=True)

        df = df[["open", "high", "low", "close", "volume", "value"]]
        df = df.astype({"open": int, "high": int, "low": int, "close": int, "volume": int, "value": int})

        return df


def get_ohlcv_by_codes_for_period(
    codes: List[str],
    start_date: datetime.date,
    end_date: Optional[datetime.date] = None,
    adjusted: bool = True,
    ascending: bool = False,
    exchange: Union[str, DataExchange] = "KRX",
) -> Dict[str, pd.DataFrame]:
    """
    지정된 코드 리스트와 기간에 대한 OHLCV 데이터를 조회합니다.

    이 함수는 하나 이상의 주식 코드와 시작 날짜, 선택적으로 종료 날짜를 지정하여 해당 기간 동안의 OHLCV 데이터를 API를 통해 요청합니다.
    반환된 데이터는 각 주식 코드별로 시가, 고가, 저가, 종가, 거래량을 포함하는 pandas DataFrame 객체들로 구성된 딕셔너리 형태로 제공됩니다.
    각 DataFrame은 해당 주식의 날짜를 기반으로 역순으로 정렬됩니다.

    KRX: 2018년 1월 1일 데이터 부터 조회 가능합니다.
    NXT: 2025년 3월 4일 데이터 부터 조회 가능합니다.

    Args:
        codes (List[str]): 조회할 주식 코드들의 리스트.
        start_date (datetime.date): 조회할 기간의 시작 날짜.
        end_date (Optional[datetime.date]): 조회할 기간의 종료 날짜. 지정하지 않으면 최근 거래일 까지 조회됩니다.
        adjusted (bool): 수정주가 여부. 기본값은 True.
        ascending (bool): 날짜 오름차순 여부. 기본값은 False.
        exchange (Union[str, DataExchange]): 거래소. 기본값은 KRX.

    Returns:
        dict: 주식 코드를 키로 하고, 해당 코드의 OHLCV 데이터를 포함하는 DataFrame을 값으로 하는 딕셔너리.

        DataFrame의 컬럼은 다음과 같습니다.

        - open (int): 시가.
        - high (int): 고가.
        - low (int): 저가.
        - close (int): 종가.
        - volume (int): 거래량.
        - value (int): 거래대금.

    Raises:
        HTTPError: API 요청이 실패했을 때 발생.

    Examples:
        >>> dfs = get_ohlcv_by_codes_for_period(['005930', '319640'], datetime.date(2025, 7, 28), datetime.date(2025, 7, 30))
        >>> print(dfs)
        {'319640':              open   high    low  close  volume      value
                    date
                    2025-07-30  20860  20930  20840  20870   47431  990834244
                    2025-07-29  20795  20850  20765  20810   39627  824529926
                    2025-07-28  20900  20985  20880  20965   26610  556764707,
        '005930':              open   high    low  close    volume          value
                    date
                    2025-07-30  71000  73700  70600  72600  34761444  2521446314798
                    2025-07-29  70800  70800  68800  70600  28190940  1977721049950
                    2025-07-28  68200  70400  67200  70400  35332500  2431396606750}
    """
    exchange = DataExchange.validate(exchange)
    tz = pytz.timezone("Asia/Seoul")
    chunks = chunk(codes, 20)
    result = {}

    for i, asset_codes in enumerate(chunks):
        url = f"{c.PYQQQ_API_URL}/domestic-stock/ohlcv/daily/series"
        params = {
            "codes": ",".join(asset_codes),
            "start_date": start_date,
            "adjusted": "true" if adjusted else "false",
            "current_date": datetime.date.today(),
            "exchange": exchange.value,
        }
        if end_date is not None:
            params["end_date"] = end_date

        r = send_request("GET", url, params=params)
        raise_for_status(r)

        data = r.json()
        cols = data["cols"]
        dataset = data["rows"]

        if not dataset:
            continue

        for code in dataset.keys():
            rows = dataset[code]
            for row in rows:
                dt = row[0]
                if dt[-1] == "Z":
                    dt = dt[:-1] + "+00:00"
                dt = datetime.datetime.fromisoformat(dt).astimezone(tz).replace(tzinfo=None)
                row[0] = dt

            rows.reverse()

            df = pd.DataFrame(rows, columns=cols)

            # NOTE 서버 응답에서 diff, diff_rate 가 제거될 예정
            columns_to_drop = ["diff", "diff_rate"]
            df.drop(columns=[col for col in columns_to_drop if col in df.columns], inplace=True)
            df.set_index("date", inplace=True)

            df = df[["open", "high", "low", "close", "volume", "value"]]
            df = df.astype({"open": int, "high": int, "low": int, "close": int, "volume": int, "value": int})
            df.sort_index(ascending=ascending, inplace=True)
            result[code] = df

    return result



import time


_MAX_CALLS = 5
_PERIOD = 1
_SCOPE = "default"


class CallLimiter:
    """
    API 호출 제한을 관리하기 위한 싱글턴 클래스입니다.

    이 클래스는 API 호출 빈도를 제한하여 너무 많은 요청으로 인해 서비스 제한을 받는 것을 방지합니다.
    이 클래스의 인스턴스는 전역에서 단 하나만 존재하며, 다양한 'scope'에 대한 호출 제한 윈도우를 관리합니다.
    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(CallLimiter, cls).__new__(cls)
            cls._instance.windows = {"default": []}

        return cls._instance

    def wait_limit_rate(self, max_calls: int = _MAX_CALLS, period: float = _PERIOD, scope: str = _SCOPE):
        """
        지정된 시간 간격 동안 최대 호출 횟수를 초과하지 않도록 기다립니다.

        호출 윈도우를 관리하여 max_calls 및 period를 기반으로 호출이 허용될 때까지 대기합니다.
        'scope'에 따라 다른 API 또는 함수 호출에 대해 다른 제한을 설정할 수 있습니다.

        Args:
            max_calls (int): 지정된 시간 동안 허용되는 최대 호출 횟수.
            period (float): 호출 제한을 적용할 시간 간격(초).
            scope (str): 호출 제한을 적용할 범위 또는 이름.

        Examples:
            >>> call_limiter = CallLimiter()
            >>> call_limiter.wait_limit_rate(10, 60, 'api_calls')
        """
        if scope not in self.windows:
            self.windows[scope] = []

        now = time.time()
        window = list(filter(lambda x: now - x < period, self.windows[scope]))
        window_size = len(window)

        if window_size >= max_calls:
            time.sleep(period)
        elif window_size > 0:
            time.sleep(period / max_calls)

        window.append(time.time())
        self.windows[scope] = window


def limit_calls(max_calls: int = _MAX_CALLS, period: float = _PERIOD, scope: str = _SCOPE):
    """
    여러 함수에 대한 호출을 제한하는 데코레이터입니다.

    Args:
        max_calls (int): 최대 호출 횟수
        period (float): 기간(초) 동안의 호출 횟수를 제한합니다.
        scope (str): 호출 제한을 적용할 스코프 이름입니다.
    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            CallLimiter().wait_limit_rate(max_calls=max_calls, period=period, scope=scope)
            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator
